//------------------------------------------------------------------------------
//
//  npgl.c
//
//  ANTz application based on NPE is free software
//	NPE is the Neural Physics Engine that drives the interactions
//
//	ANTz hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
//	Copyright (c) 2010-2011, Shane Saxon
//	All rights reserved.
//	Redistribution and use in source and binary forms, with or without 
//	modification, are permitted provided that the following conditions are met:
//	-	Redistributions of source code must retain the above copyright notice, 
//		this list of conditions and the following disclaimer.
//	-	Redistributions in binary form must reproduce the above copyright 
//		notice, this list of conditions and the following disclaimer in the 
//		documentation and/or other materials provided with the distribution.
//	-	Neither the name of the Shane Saxon nor the names of its contributors 
//		may be used to endorse or promote products derived from this software 
//		without specific prior written permission.
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//	POSSIBILITY OF SUCH DAMAGE.
//
//------------------------------------------------------------------------------

#include "npgl.h"

#include "gl/soil/src/SOIL.h"

#include "../npctrl.h"
#include "../os/npos.h"

#include "watchtools.h"


unsigned int texture;									// texture, debug zz


void npGLResizeScene (int width, int height);
void npCopyGPUtoCPU (void* dataRef);


void npDrawTextLabels();										//zz-CAMERA

void npPick( int x, int y );											//debug, zz //MB//

//------------------------------------------------------------------------------
void npInitGL(void* dataRef) 
{
	npInitGLDraw (dataRef);

	npInitGLDraw (dataRef);

//	glShadeModel (GL_FLAT);
//	glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
//	glClearDepth (1.0f);

	glEnable (GL_DEPTH_TEST);

//	glDepthFunc (GL_LEQUAL);
//	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	// load our texture												// texture, debug zz
	texture = SOIL_load_OGL_texture
	(
		"map00001.jpg",
		SOIL_LOAD_AUTO,
		SOIL_CREATE_NEW_ID,
		SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y
		| SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
	);																	// texture, debug zz
}


// currently just single window support, add multi-screen capability later... debug zz
// redbook has info on how to calculate the projections
//------------------------------------------------------------------------------
void npGLResizeScene (int wd, int ht)							
{
	int camIndex = 1;	//update with list in data->io->view.camIndex[]		debug zz
	GLfloat ratio = 1.0;

	pData data = npGetDataRef();

	NPnodePtr camNode = data->map.node[camIndex];
	NPcameraPtr camData = camNode->data;

	//MB//						//was width, made global by MB, undo it... debug, zz
	width  = wd;
	height = ht;
	//MB-end//

	if (height > 0)							// Prevent a divide by zero
		camData->aspectRatio = (GLfloat)width / (GLfloat)height;

	// Reset the coordinate system before modifying
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	
	// Set the viewport to be the entire window
    glViewport(0, 0, width, height);

	// Set the clipping volume
	gluPerspective (camData->fov, camData->aspectRatio, 
					camData->clipNear, camData->clipFar);
	glMatrixMode (GL_MODELVIEW);

	npGLDrawScene (data);
}


//------------------------------------------------------------------------------
void npGLDrawScene (void* dataRef)
{
	//MB-LABEL
	NPfloatXYZ	trans;
	NPfloatXYZA rot;
	GLfloat uX, uY, uZ;
	//MB-END

	int	camIndex = 1;
	GLfloat upX = 0.0f, upY = 0.0f, upZ = 1.0f;
	GLfloat angle = 0.0;

	GLfloat position[] = { -150.0f, 150.0f, -150.0f, 1.0f };
	GLfloat positionB[] = { 150.0f, -150.0f, 150.0f, 1.0f };
	GLfloat positionC[] = { 0.0f, 0.0f, 15.0f, 1.0f };

	GLfloat globalAmbient[] = { 0.3f, 0.3f, 0.3f, 1.0f };

	GLfloat materialSpecular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
	GLfloat materialShininess[] = { 128.0f };
	GLfloat materialAmbient[] = { 0.25f, 0.25f, 0.25f, 1.0f };
	GLfloat materialDiffuse[] = { 0.4f, 0.4f, 0.4f, 1.0f };
	GLfloat localAmbient[] = { 0.0f, 0.0f, 0.0f, 1.0f };

	GLfloat redDiffuseMaterial[] = {1.0f, 1.0f, 1.0f}; //set the material to red
	GLfloat whiteSpecularMaterial[] = {1.0f, 1.0f, 1.0f}; //set the material to white
	GLfloat greenEmissiveMaterial[] = {0.1f, 0.1f, 0.1f}; //set the material to green
	GLfloat whiteSpecularLight[] = {1.0f, 1.0f, 1.0f}; //set the light specular to white
	GLfloat blackAmbientLight[] = {0.0f, 0.0f, 0.0f}; //set the light ambient to black
	GLfloat whiteDiffuseLight[] = {1.0f, 1.0f, 1.0f}; //set the diffuse light to white
	GLfloat blankMaterial[] = {0.0f, 0.0f, 0.0f}; //set the diffuse light to white
	GLfloat mShininess[] = {128}; //set the shininess of the material

	GLfloat redNotGreenNotBlue[] = {1.0f, -1.0f, -1.0f}; //set the material to red

	pData data = (pData) dataRef;
	NPnodePtr camNode = cameraSelected();
	NPcameraPtr camData = camNode->data;

	NPnodePtr gridNode = data->map.node[2];					// texture, debug zz
	NPgridPtr gridData = gridNode->data;

	GLfloat gridX = 0.0f, gridY = 0.0f;

	//lighting
//	glLightModelfv (GL_LIGHT_MODEL_AMBIENT, globalAmbient);
	
	npUpdateCtrl (dataRef);	  //update data, positions, physics, interactions...

	glClearColor (data->io.clear.r, data->io.clear.g, data->io.clear.b,
					data->io.clear.a);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//	glfwSwapInterval(3);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
	glEnable (GL_LIGHT1);
//	glEnable (GL_LIGHT2);		//MB-Transp

	glLightfv (GL_LIGHT0, GL_POSITION, position);
	glLightfv(GL_LIGHT0, GL_SPECULAR, whiteSpecularLight);
    glLightfv(GL_LIGHT0, GL_AMBIENT, blackAmbientLight);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteDiffuseLight);

	glLightfv (GL_LIGHT1, GL_POSITION, positionB);
	glLightfv(GL_LIGHT1, GL_SPECULAR, whiteSpecularLight);
    glLightfv(GL_LIGHT1, GL_AMBIENT, blackAmbientLight);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, whiteDiffuseLight);

//	glLightfv (GL_LIGHT2, GL_POSITION, positionC);
//	glLightfv(GL_LIGHT2, GL_SPECULAR, redNotGreenNotBlue);
//   glLightfv(GL_LIGHT2, GL_AMBIENT, blackAmbientLight);
//    glLightfv(GL_LIGHT2, GL_DIFFUSE, redNotGreenNotBlue);

//	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, whiteSpecularMaterial);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mShininess);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, redDiffuseMaterial);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, greenEmissiveMaterial);

	glEnable (GL_COLOR_MATERIAL);


	//set the alphamode
	if (data->io.gl.alphaMode)
	{
		switch (data->io.gl.alphaMode)
		{
			case kNPalphaModeSubtractive :
				glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				break;
			case kNPalphaModeAdditive :
				glBlendFunc (GL_SRC_COLOR, GL_ONE_MINUS_SRC_ALPHA);
				break;
			case kNPalphaModeDark :
				glBlendFunc (GL_SRC_ALPHA, GL_DST_COLOR);
				break;
			default :
				glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	//MB-Transp
				break;
		}
		glEnable (GL_BLEND);
	}
	else
		glDisable (GL_BLEND);


	glEnable (GL_DEPTH_TEST);

	//set shading for flat or smooth
	if (data->io.gl.shade == kNPglShadeSmooth)
		glShadeModel (GL_SMOOTH);
	else
		glShadeModel (GL_FLAT);


	glEnable (GL_CULL_FACE);		//do not render back-faces

	//recalculate normals for non-uniform scaling or rescale normals for speed
	switch (data->io.gl.normal)
	{
		case kNPglNormalize :				//geo-computationally slow
	//		glDisable (RESCALE_NORMAL);
			glEnable (GL_NORMALIZE);	//recalc normals for non-uniform scaling
			break;
		case kNPglRescaleNormal :			//fast								//debug, zz
			glDisable (GL_NORMALIZE);
			data->io.gl.normal++;			//workaround until RESCALE_NORMAL fixed, zz
	//		glEnable (RESCALE_NORMAL);	//rescale normals, assumes uniform scaling
			break;
		default :							//fastest
			glDisable (GL_NORMALIZE);
	//		glDisable (RESCALE_NORMAL);
			break;
	}

    // setup texture mapping									// texture, debug zz
	gridX = gridNode->scale.x * 5.0f * (GLfloat) gridData->segments.x;
	gridY = gridNode->scale.y * 5.0f * (GLfloat) gridData->segments.y;

	if (!npGetPickPass())	//allow for click on background tex's for mouse flying
	{
	 glDisable (GL_LIGHTING);
      glEnable( GL_TEXTURE_2D );
      glBindTexture( GL_TEXTURE_2D, texture );

	  glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
      glPushMatrix();
      glRotatef( 0.0f, 0.0f, 0.0f, 1.0f );
      glBegin( GL_QUADS );
      glTexCoord2d(0.0,0.0); glVertex3d(-gridX,-gridY,-0.1f);
      glTexCoord2d(1.0,0.0); glVertex3d(+gridX,-gridY,-0.1f);
      glTexCoord2d(1.0,1.0); glVertex3d(+gridX,+gridY,-0.1f);
      glTexCoord2d(0.0,1.0); glVertex3d(-gridX,+gridY,-0.1f);
      glEnd();
      glPopMatrix();
	   glDisable( GL_TEXTURE_2D );
	 glEnable (GL_LIGHTING);										//end texture, debug zz
	}
	
	npDrawNodes (dataRef);				//draws mesh, graphs, pins, etc...
	
//	npCameraUpdate( dataRef );

	glLoadIdentity();
	//MB-LABEL
	
	trans.x = 0.0f;	
	trans.y = 0.0f;
	trans.z = 100.0f;
	rot.x	= 0.0f;
	rot.y	= 0.0f;
	rot.z	= -1.0f;
	rot.angle = 0.0f;
	uX		= 0.0f;
	uY		= 0.0f;
	uZ		= 1.0f;
	
#define BUGFIX_MATRIX 0.0000000001
	//HUD LookAt Camera
	glPushMatrix();
	gluLookAt (trans.x, 
			   trans.y,
			   trans.z, 
			   trans.x + rot.x,
			   trans.x + rot.y + BUGFIX_MATRIX,
			   trans.z + rot.z,
			   uX, uY, uZ );
//	npDrawTextLabels();				// draws overlays, grids, text... //MB-LABEL
	glPopMatrix();
	
	
	//3D Camera			//MB-LABEL-END
	gluLookAt (	camNode->translate.x, 
				camNode->translate.y,
				camNode->translate.z, 
				camNode->translate.x + camNode->rotate.x,
				camNode->translate.y + camNode->rotate.y,
				camNode->translate.z + camNode->rotate.z,
				upX, upY, upZ );
	
//	glOrtho ( -2.0f, 2.0f, -2.0f, 2.0f, 0.05f, 2.0f);
//	gluOrtho2D ( -2.0f, 2.0f, -2.0f, 2.0f);
	
	npDrawHUD (dataRef);				// draws overlays, grids, text...
	
/*																			//debug zz
	setOrthographicProjection();

	glPushMatrix();
	glLoadIdentity();
	renderBitmapString(5,30,GLUT_BITMAP_HELVETICA_18,"Lighthouse3D");
	glPopMatrix();

	restorePerspectiveProjection();
*/
	
	npCopyGPUtoCPU (dataRef);			// copies display to main CPU memory
	
//	glfwSleep (0.1f);
}


// copy the rasterized display to the main system memory (CPU RAM or Drive)
//------------------------------------------------------------------------------
void npCopyGPUtoCPU (void* dataRef)
{
	return;
}


//MB-CAMERA
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

/*		//zz-CAMERA
//------------------------------------------------------------------------------
void npPick (int x, int y)
{
	GLubyte r,g,b,a;	//Unsigned Byte 8-bit Range: 0-255 // changed from u_int8_t, zz debug
	char pixels[4];		//4 Bytes
	int idInRGB;

	NPnodePtr node = NULL;
	pData data = (pData) npGetDataRef();;

//	WATCH_FUNC;
//	WATCH_INT( x );

	glReadPixels( x,y, 1,1, GL_RGBA, GL_UNSIGNED_BYTE, pixels );			//Read the Pixel
	
	r = pixels[0];
	g = pixels[1];
	b = pixels[2];
	a = pixels[3];

	idInRGB = RGBtoID( r,g,b );
	selectedID = idInRGB;


	if( selectedID == 0 )
	{
		data->io.picking = false;

		// if camera not selected then select it
		if (data->map.nodeRootIndex != kNPcameraRoot)
		{
			npCtrlCommand (kGraphSelectAllOff, data);	//zz debug not efficient
			npCtrlCommand (kGraphCamera, data);
		}
	}
	else
	{
		data->io.picking = true;

		if (selectedID >= kNPnodeMax)	//mouse selection crashing workaround, debug zz
			selectedID = 0;

		node = data->map.nodeID[selectedID];
		if (node != NULL)
		{
			if (node->recordID)
				printf ("recordID: %d \n", node->recordID);
			else
				WATCH_INT( selectedID );
		}
	}
}
*/


//------------------------------------------------------------------------------
//MB-CAMERA
void npPick( int x, int y )
{
//	char* cmd[256];
	GLubyte r,g,b,a;	//Unsigned Byte 8-bit Range: 0-255 // changed from u_int8_t, zz debug
	char pixels[4];		//4 Bytes
	int idInRGB;
	
	NPnodePtr node = NULL;
	pData data = (pData) npGetDataRef();;
	
	//	WATCH_FUNC;
	//	WATCH_INT( x );
	
	glReadPixels( x,y, 1,1, GL_RGBA, GL_UNSIGNED_BYTE, pixels );			//Read the Pixel
	
	r = pixels[0];
	g = pixels[1];
	b = pixels[2];
	a = pixels[3];
	
	idInRGB = RGBtoID( r,g,b );
	selectedID = idInRGB;
	
	
	if( selectedID == 0 )
		data->io.picking = false;
	else
	{
		data->io.picking = true;
		
		if (selectedID >= kNPnodeMax)	//mouse selection crashing workaround, debug zz
			selectedID = 0;
		
		node = data->map.nodeID[selectedID];
		if (node != NULL)
		{
			if (data->io.mouse.buttonR)
				node->selected = 1 - node->selected;	//toggle value
			else
			{
				//if node not selected then clear selection, then select
				if (!node->selected)
					npCtrlCommand (kGraphSelectAllOff, data);

				node->selected = true;
			}

			data->map.selectedNode = node;
			data->map.activeNode = node;

			if (node->recordID)
			{
				printf ("recordID: %d \n", node->recordID);
				//	npDatabaseCommand("DESCRIBE stats", data); // Do string processing to append recordID to end with WHERE statement
				//	strcat(cmd, node->recordID);
//				sprintf(cmd, "SELECT * FROM stats WHERE ID = %d", node->recordID);
//				printf("sprintf");
//				npDatabaseCommand(cmd, data); // Do string processing to append recordID to end with WHERE statement
			}
			else
				WATCH_INT( selectedID );
		}
	}
	
	//MB-CAMERA
	if( data->io.picking )
	{
		data->io.mouse.pickedOnThisFrame = true;
	}
	
	//MB-CAMERA
}
//MB-





//MB-LABEL

//Draw "Ring" out of Toroid
void drawRing()
{
//	glScalef( 12,12,12 );
	
	glScalef(100,100,100);
	DrawSimpleTorus();
}


//Project CURRENT Matrix state World position Coordinates onto the Screen
NPfloatXY npProjectWorldToScreen()
{	
	float pinOffset = 6.0f;

	GLint viewport[4];

	GLdouble screenX, screenY, dummyZ;	//Screen Coordinates Output

	NPfloatXY screenXY;

	GLdouble model_matrix[16];
	GLdouble projection_matrix[16];

	//Model Matrix Current
	glGetDoublev(GL_MODELVIEW_MATRIX, model_matrix);
	
	//Projection Matrix Current
	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);

	//Viewport
	glGetIntegerv(GL_VIEWPORT, viewport);	

	//Do the projection
	gluProject( 0,0,pinOffset, model_matrix, projection_matrix, viewport, &screenX, &screenY, &dummyZ );
	
//	WATCH_FLOAT( screenX );
//	WATCH_FLOAT( screenY );
//	WATCH_FLOAT( dummyZ  );

	screenXY.x = (float)screenX;
	screenXY.y = (float)screenY;

	return screenXY;
}

//Draw the Text Labels (Simple Ring now)
void npDrawTextLabels()
{
	int i = 0;
	float sc;
	int nodeRootCount;
				
	NPfloatXY screenPosXY;

	pData data = (pData) npGetDataRef();
	NPnodePtr node;


//	WATCH_FUNC;
	
	glPushMatrix();
	
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glColor4f( 1,1,1,  1 );
	
	sc = 62.5f/height;
	glScalef( sc, sc, sc );		//scale to Screen Coords
	
	nodeRootCount = data->map.nodeRootCount;
//	WATCH_INT( nodeRootCount );
	
	for( i = 0; i < nodeRootCount; i++ )
	{
		if( i < 3 )	//Nothing, Camera, Grid
			continue;
		
		node = data->map.node[i];
		
		screenPosXY = node->screenXY;
//		WATCH_VEC2( screenPosXY );
		screenPosXY.x -= width / 2.0f;
		screenPosXY.y -= height/ 2.0f;
		
		glPushMatrix();
			glTranslated( screenPosXY.x,screenPosXY.y,0 );	
			drawRing();
		glPopMatrix();
	}
	
	glPopMatrix();
}

